﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using System.Xml;
using System.Diagnostics;
using HeuristicLab.DataAnalysis;
using HeuristicLab.Operators;
using HeuristicLab.Random;
using HeuristicLab.Selection;
using HeuristicLab.Logging;
using HeuristicLab.Data;
using HeuristicLab.Operators.Programmable;

namespace HeuristicLab.GP.StructureIdentification {
  public class StandardGP : AlgorithmBase, IEditable {

    private IntData maxGenerations = new IntData();
    public int MaxGenerations {
      get { return maxGenerations.Data; }
      set { maxGenerations.Data = value; }
    }

    private IntData tournamentSize = new IntData();
    public int TournamentSize {
      get { return tournamentSize.Data; }
      set { tournamentSize.Data = value; }
    }

    private DoubleData fullTreeShakingFactor = new DoubleData();
    public double FullTreeShakingFactor {
      get { return fullTreeShakingFactor.Data; }
      set { fullTreeShakingFactor.Data = value; }
    }

    private DoubleData onepointShakingFactor = new DoubleData();
    public double OnePointShakingFactor {
      get { return onepointShakingFactor.Data; }
      set { onepointShakingFactor.Data = value; }
    }

    public override int PopulationSize {
      get {
        return base.PopulationSize;
      }
      set {
        base.PopulationSize = value;
        Parents = 2 * value;
      }
    }

    public StandardGP()
      : base() {
      PopulationSize = 10000;
      MaxGenerations = 100;
      TournamentSize = 7;
      MutationRate = 0.15;
      Elites = 1;
      MaxTreeSize = 100;
      MaxTreeHeight = 10;
      FullTreeShakingFactor = 0.1;
      OnePointShakingFactor = 1.0;

      Engine = new SequentialEngine.SequentialEngine();
      IOperator algo = CreateAlgorithm();
      Engine.OperatorGraph.AddOperator(algo);
      Engine.OperatorGraph.InitialOperator = algo;
    }

    internal override IOperator CreateSelector() {
      TournamentSelector selector = new TournamentSelector();
      selector.Name = "Selector";
      selector.GetVariableInfo("Selected").ActualName = "Parents";
      selector.GetVariableInfo("GroupSize").Local = false;
      selector.RemoveVariable("GroupSize");
      selector.GetVariableInfo("GroupSize").ActualName = "TournamentSize";
      return selector;
    }

    internal override IOperator CreateGlobalInjector() {
      VariableInjector globalInjector = (VariableInjector)base.CreateGlobalInjector();
      globalInjector.AddVariable(new HeuristicLab.Core.Variable("TournamentSize", tournamentSize));
      globalInjector.AddVariable(new HeuristicLab.Core.Variable("MaxGenerations", maxGenerations));
      return globalInjector;
    }

    internal override IOperator CreateCrossover() {
      StandardCrossOver crossover = new StandardCrossOver();
      crossover.Name = "Crossover";
      crossover.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";
      return crossover;
    }

    internal override IOperator CreateTreeCreator() {
      ProbabilisticTreeCreator treeCreator = new ProbabilisticTreeCreator();
      treeCreator.Name = "Tree generator";
      treeCreator.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";
      treeCreator.GetVariableInfo("MinTreeSize").Local = true;
      treeCreator.AddVariable(new HeuristicLab.Core.Variable("MinTreeSize", new IntData(30)));
      return treeCreator;
    }

    internal override IOperator CreateFunctionLibraryInjector() {
      return new FunctionLibraryInjector();
    }

    internal override IOperator CreateManipulator() {
      CombinedOperator manipulator = new CombinedOperator();
      manipulator.Name = "Manipulator";
      StochasticMultiBranch multibranch = new StochasticMultiBranch();
      FullTreeShaker fullTreeShaker = new FullTreeShaker();
      fullTreeShaker.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";
      fullTreeShaker.GetVariableInfo("ShakingFactor").Local = true;
      fullTreeShaker.AddVariable(new HeuristicLab.Core.Variable("ShakingFactor", fullTreeShakingFactor));

      OnePointShaker onepointShaker = new OnePointShaker();
      onepointShaker.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";
      onepointShaker.GetVariableInfo("ShakingFactor").Local = true;
      onepointShaker.AddVariable(new HeuristicLab.Core.Variable("ShakingFactor", onepointShakingFactor));
      ChangeNodeTypeManipulation changeNodeTypeManipulation = new ChangeNodeTypeManipulation();
      changeNodeTypeManipulation.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";
      CutOutNodeManipulation cutOutNodeManipulation = new CutOutNodeManipulation();
      cutOutNodeManipulation.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";
      DeleteSubTreeManipulation deleteSubTreeManipulation = new DeleteSubTreeManipulation();
      deleteSubTreeManipulation.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";
      SubstituteSubTreeManipulation substituteSubTreeManipulation = new SubstituteSubTreeManipulation();
      substituteSubTreeManipulation.GetVariableInfo("OperatorLibrary").ActualName = "FunctionLibrary";

      IOperator[] manipulators = new IOperator[] {
        onepointShaker, fullTreeShaker,
        changeNodeTypeManipulation,
        cutOutNodeManipulation,
        deleteSubTreeManipulation,
        substituteSubTreeManipulation};

      DoubleArrayData probabilities = new DoubleArrayData(new double[manipulators.Length]);
      for (int i = 0; i < manipulators.Length; i++) {
        probabilities.Data[i] = 1.0;
        multibranch.AddSubOperator(manipulators[i]);
      }
      multibranch.GetVariableInfo("Probabilities").Local = true;
      multibranch.AddVariable(new HeuristicLab.Core.Variable("Probabilities", probabilities));

      manipulator.OperatorGraph.AddOperator(multibranch);
      manipulator.OperatorGraph.InitialOperator = multibranch;
      return manipulator;
    }

    internal override IOperator CreateBestSolutionProcessor() {
      SequentialProcessor bestSolutionProcessor = new SequentialProcessor();
      MeanAbsolutePercentageErrorEvaluator trainingMapeEvaluator = new MeanAbsolutePercentageErrorEvaluator();
      trainingMapeEvaluator.Name = "TrainingMapeEvaluator";
      trainingMapeEvaluator.GetVariableInfo("MAPE").ActualName = "TrainingMAPE";
      trainingMapeEvaluator.GetVariableInfo("SamplesStart").ActualName = "TrainingSamplesStart";
      trainingMapeEvaluator.GetVariableInfo("SamplesEnd").ActualName = "TrainingSamplesEnd";
      MeanAbsolutePercentageErrorEvaluator validationMapeEvaluator = new MeanAbsolutePercentageErrorEvaluator();
      validationMapeEvaluator.Name = "ValidationMapeEvaluator";
      validationMapeEvaluator.GetVariableInfo("MAPE").ActualName = "ValidationMAPE";
      validationMapeEvaluator.GetVariableInfo("SamplesStart").ActualName = "ValidationSamplesStart";
      validationMapeEvaluator.GetVariableInfo("SamplesEnd").ActualName = "ValidationSamplesEnd";
      ProgrammableOperator progOperator = new ProgrammableOperator();
      progOperator.RemoveVariableInfo("Result");
      progOperator.AddVariableInfo(new HeuristicLab.Core.VariableInfo("EvaluatedSolutions", "", typeof(IntData), VariableKind.In));
      progOperator.Code = @"
int evalSolutions = EvaluatedSolutions.Data;
scope.AddVariable(new Variable(""EvaluatedSolutions"", new IntData(evalSolutions)));
";
      bestSolutionProcessor.AddSubOperator(trainingMapeEvaluator);
      bestSolutionProcessor.AddSubOperator(validationMapeEvaluator);
      bestSolutionProcessor.AddSubOperator(progOperator);
      return bestSolutionProcessor;
    }

    public virtual IEditor CreateEditor() {
      return new StandardGpEditor(this);
    }

    public override IView CreateView() {
      return new StandardGpEditor(this);
    }

    public override object Clone(IDictionary<Guid, object> clonedObjects) {
      StandardGP clone = new StandardGP();
      clonedObjects.Add(Guid, clone);
      clone.Engine = (SequentialEngine.SequentialEngine)Auxiliary.Clone(Engine, clonedObjects);
      return clone;
    }

    #region SetReferences Method
    private void SetReferences() {
      // SGA
      CombinedOperator co1 = (CombinedOperator)Engine.OperatorGraph.InitialOperator;
      // SequentialProcessor in SGA
      SequentialProcessor algorithm = (SequentialProcessor)co1.OperatorGraph.InitialOperator;
      // RandomInjector
      RandomInjector ri = (RandomInjector)algorithm.SubOperators[1];
      // VariableInjector
      VariableInjector vi = (VariableInjector)algorithm.SubOperators[2];
      maxGenerations = vi.GetVariable("MaxGenerations").GetValue<IntData>();
      tournamentSize = vi.GetVariable("TournamentSize").GetValue<IntData>();
    }
    #endregion

    #region Persistence Methods
    public override XmlNode GetXmlNode(string name, XmlDocument document, IDictionary<Guid, IStorable> persistedObjects) {
      XmlNode node = base.GetXmlNode(name, document, persistedObjects);
      node.AppendChild(PersistenceManager.Persist("Engine", Engine, document, persistedObjects));
      return node;
    }
    public override void Populate(XmlNode node, IDictionary<Guid, IStorable> restoredObjects) {
      base.Populate(node, restoredObjects);
      Engine = (SequentialEngine.SequentialEngine)PersistenceManager.Restore(node.SelectSingleNode("Engine"), restoredObjects);
      SetReferences();
    }
    #endregion
  }
}
